TASK = "task"
FRAMEWORK = "framework"
RESULT = "result"
DURATION = "duration"
PREDICT_DURATION = "predict_duration"

TIME_TRAIN_S = "time_train_s"
TIME_INFER_S = "time_infer_s"

DATASET = "dataset"
FOLD = "fold"
RANK = "rank"

METRIC_ERROR = "metric_error"
METRIC_SCORE = "metric_score"
METRIC = "metric"
PROBLEM_TYPE = "problem_type"

NUM_CLASSES = "num_classes"
NUM_COLS = "num_cols"
NUM_ROWS = "num_rows"
DATETIME = "datetime"

ERROR_COUNT = "error_count"
RANK_1 = "rank=1_count"

BESTDIFF = "bestdiff"

LOSS_RESCALED = "loss_rescaled"
