metrics = [
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Training",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Prediction",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Training",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "credit-g",
        "mode": "Prediction",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Training",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Prediction",
        "fold": 0,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Training",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
    {
        "framework": "AutoGluon",
        "dataset": "vehicle",
        "mode": "Prediction",
        "fold": 1,
        "metric": "CPUUtilization",
        "statistic_type": "Average",
        "statistic_value": 11.472356376239336,
        "unit": "Percent",
    },
]
