_TEXT = "text"
_NER = "ner"
_CATEGORICAL = "categorical"
_NUMERICAL = "numerical"
_BINARY = "binary"
_MULTICLASS = "multiclass"
_REGRESSION = "regression"
_IMAGE_SIMILARITY = "image_similarity"
_IMAGE_TEXT_SIMILARITY = "image_text_similarity"
_TEXT_SIMILARITY = "text_similarity"
_OBJECT_DETECTION = "object_detection"
_FEW_SHOT_CLASSIFICATION = "few_shot_classification"
